pub mod mesh;
pub mod interop;

// Importing Bevy ECS components and systems
use crate::mesh::create_rectangle_mesh;
use bevy::ecs::component::Component;
use bevy::ecs::system::{Commands, Query};
use bevy::ecs::entity::Entity;
use bevy::ecs::query::Without;
use bevy::ecs::system::ResMut;
use crate::interop::{ToPoint3, ToVec3};

// Importing Bevy assets and rendering related components
use bevy::prelude::*;
use bevy::render::mesh::{Indices, PrimitiveTopology};
use bevy::asset::Assets;
use bevy::pbr::{PbrBundle, StandardMaterial};
use bevy::render::color::Color;
use bevy::transform::components::Transform;
use bevy::utils::default;

// Importing Bevy application related components
use bevy::app::{App, Plugin, Update};

// Importing Fj-core functionalities

use fj_core::algorithms::triangulate::Triangulate;
use fj_core::objects::{Sketch, Solid};
use fj_core::storage::Handle as FjHandle;

// Importing Fj-interop mesh
use fj_interop::mesh::Mesh as FjMesh;

// Importing Fj-math and other standard functionalities

use fj_math::{Aabb, Line, Point, Scalar, Vector};
use std::ops::Deref;
use bevy::a11y::accesskit::Size;
use bevy::math::Vec3;
use bevy::ui::ZIndex::Global;
use fj_core::geometry::{GlobalPath, SurfacePath};
use fj_core::objects::Object::Surface;
use fj_core::operations::holes::AddHole;
use fj_core::operations::split::{SplitEdge, SplitFace, SplitHalfEdge};
use rand::Rng;
use crate::interop::FjMeshWrapper;
use crate::mesh::{convert_mesh, LineList};


#[derive(Component)]
struct FjConvertedFlag;

pub struct FjRenderPlugin;

impl Plugin for FjRenderPlugin {
    fn build(&self, app: &mut App) {
        app.add_systems(Update, update_fj_model_system);
    }
}

fn update_fj_model_system(
    mut commands: Commands,
    query: Query<(Entity, &FjMeshWrapper), Without<FjConvertedFlag>>,
    mut meshes: ResMut<Assets<Mesh>>,
    mut materials: ResMut<Assets<StandardMaterial>>,
    asset_server: Res<AssetServer>
) {

    for (entity, solid) in &query {

        let bevy_mesh = convert_mesh(&solid.mesh);
        commands.entity(entity).insert(
            (
                PbrBundle {
                    mesh: meshes.add(Mesh::from(bevy_mesh.clone())),
                    material: materials.add(Color::rgb(0.8, 0.7, 0.6).into()),
                    transform: Transform::from_xyz(0.0, 0.0, 0.0),
                    ..default()
                },
                FjConvertedFlag
            ));
        add_debug_info_to_entity(&mut commands, solid, &mut meshes, &mut materials);
    }
}


fn add_debug_info_to_entity(mut commands: &mut Commands,
                            fj_model: &FjMeshWrapper,
                            mut meshes: &mut ResMut<Assets<Mesh>>,
                            mut materials: &mut ResMut<Assets<StandardMaterial>>,
) {
    // for region in fj_model.sketch.regions() {
    //     for cycle in region.all_cycles() {
    //         for half_edge in cycle.half_edges() {
    //             match half_edge.path() {
    //                 SurfacePath::Circle(x) => {}
    //                 SurfacePath::Line(x) => {
    //                     render_line(commands, fj_model, meshes, materials, x)
    //                 }
    //             }
    //             // 1d point
    //             half_edge.boundary().inner[0].coords;
    //             half_edge.boundary().inner[1].coords;
    //         }
    //     }
    //
    // }

    for shell in fj_model.handle.shells(){

    }


    for shell in fj_model.handle.shells() {
        println!("{:?}. shell", shell);
        for face in shell.faces() {

            println!("{:?}. face", face);
            let surface = face.surface();
            let geom = surface.geometry();

            let geom_sweep_vector = geom.v;

            for edge in face.region().exterior().half_edges(){
                match edge.path() {
                    SurfacePath::Circle(_) => {}
                    SurfacePath::Line(x) => {
                        let mut rng = rand::thread_rng();
                        let color = Color::rgb(1.0, 0.,0.).into();

                        let mut origin = geom.point_from_surface_coords(x.origin()).to_xyz();
                        let direction = geom.vector_from_surface_coords(x.direction()).to_xyz();
                        let opposite_corner = Point{ coords: Vector{
                            components: [
                                Scalar::from(origin.x.into_f32()) + direction.x,
                                Scalar::from(origin.y.into_f32()) + direction.y,
                                Scalar::from(origin.z.into_f32()) + direction.z,
                            ]
                        }};
                        let line = GlobalPath::line_from_points([origin, opposite_corner]);
                        match line.0 {
                            GlobalPath::Circle(_) => {}
                            GlobalPath::Line(x) => {
                                render_line(commands, fj_model, meshes, materials, x, color);
                            }
                        }
                    }
                }
            }

            match geom.u {
                GlobalPath::Circle(x) => {
                    x.aabb();
                }
                GlobalPath::Line(x) => {

                    let mut rng = rand::thread_rng();
                    let red: f32 = rng.gen_range(0.0..1.0);
                    let green: f32 = rng.gen_range(0.0..1.0);
                    let blue: f32 = rng.gen_range(0.0..1.0);
                    // let color = Color::rgb(red, green, blue).into();
                    // render_line(commands, fj_model, meshes, materials, x, color);

                    // let origin = x.origin();
                    // let direction = x.direction();
                    // let opposite_corner = Vec3::new(
                    //     origin.x.into_f32() + direction.x.into_f32(),
                    //     origin.y.into_f32() + direction.y.into_f32(),
                    //     origin.z.into_f32() + direction.z.into_f32(),
                    // );
                    //
                    // commands.spawn(( // line on top of sweep, just offset by the sweep vector
                    //                  PbrBundle {
                    //                      // mesh:  meshes.add(Mesh::from(LineList {
                    //                      //     lines: vec![
                    //                      //         (opposite_corner.into(), origin.to_vec3()),
                    //                      //     ],
                    //                      // })),
                    //                      mesh: meshes.add(create_rectangle_mesh(opposite_corner, origin.to_vec3(), 0.01, 0.01)),
                    //                      material: materials.add(Color::rgb(red, green, blue).into()),
                    //                      transform: Transform::from_xyz(geom_sweep_vector.x.into_f32(), geom_sweep_vector.y.into_f32(), geom_sweep_vector.z.into_f32()),
                    //                      ..default()
                    //                  },
                    //                  // RaycastPickTarget::default(), // <- Needed for the raycast backend.
                    //                  // PickableBundle::default() // <- This one too
                    // ));
                    // commands.spawn(( // line following the sweep
                    //                  PbrBundle {
                    //                      mesh: meshes.add(create_rectangle_mesh(origin.to_vec3() + geom_sweep_vector.to_vec3(), origin.to_vec3(), 0.01, 0.01)),
                    //                      material: materials.add(Color::rgb(red, green, blue).into()),
                    //                      transform: Transform::from_xyz(0.0, 0.0, 0.0),
                    //                      ..default()
                    //                  },
                    //                  // RaycastPickTarget::default(), // <- Needed for the raycast backend.
                    //                  // PickableBundle::default() // <- This one too
                    // ));
                    // commands.spawn(( // swept vertex
                    //                  PbrBundle {
                    //                      mesh: meshes.add(Mesh::from(shape::Cube { size: 0.025 })),
                    //                      material: materials.add(Color::rgb(1.0, 1.0, 1.0).into()),
                    //                      transform: Transform::from_xyz(
                    //                          origin.to_xyz().x.into_f32() + geom_sweep_vector.to_vec3().x,
                    //                          origin.to_xyz().y.into_f32() + geom_sweep_vector.to_vec3().y,
                    //                          origin.to_xyz().z.into_f32() + geom_sweep_vector.to_vec3().z,
                    //                      ),
                    //                      // transform: Transform::from_scale(origin.to_vec3() + geom_sweep_vector.to_vec3()),
                    //                      ..default()
                    //                  },
                    //                  // RaycastPickTarget::default(), // <- Needed for the raycast backend.
                    //                  // PickableBundle::default(), // <- This one too
                    //
                    //                  // I want to  insert a component when dragging starts, that contains
                    //                  // data about the exact ray hit location.
                    //                  // On::<Pointer<DragStart>>::run(run),
                    //                  // On::<Pointer<DragStart>>::target_insert(DragCaster::default()),
                    //                  // On::<Pointer<Drag>>::target_component_mut::<DragCaster>(|drag, mut caster| {
                    //                  //     caster.hit_location = drag.hit.position.unwrap();
                    //                  //
                    //                  //     // (*transform).translation = transform.translation + Vec3::new(-drag.delta.x / 100.0, 0.0, drag.delta.y / 100.0);
                    //                  // }),
                    //                  // On::<Pointer<Drag>>::target_component_mut::<DragCaster>(|drag, mut caster| {
                    //                  //     drag.distance
                    //                  //     // (*transform).translation = transform.translation + Vec3::new(-drag.delta.x / 100.0, 0.0, drag.delta.y / 100.0);
                    //                  // }),
                    // ));
                }
            }
        }
    }
}

fn render_line<const T: usize>(mut commands: &mut Commands,
                               fj_model: &FjMeshWrapper,
                               mut meshes: &mut ResMut<Assets<Mesh>>,
                               mut materials: &mut ResMut<Assets<StandardMaterial>>,
                               line: Line<T>,
                               color: Color) {
    let origin = line.origin().to_xyz();
    let direction = line.direction().to_xyz();
    let opposite_corner = Vec3::new(
        origin.x.into_f32() + direction.x.into_f32(),
        origin.y.into_f32() + direction.y.into_f32(),
        origin.z.into_f32() + direction.z.into_f32(),
    );

    commands.spawn(( // line on base of sweep
                     PbrBundle {
                         mesh: meshes.add(Mesh::from(create_rectangle_mesh(opposite_corner, origin.to_vec3(), 0.01, 0.01))),
                         // mesh: meshes.add(Mesh::from(shape::Box::from_corners(opposite_corner.into(), origin.to_vec3()))),
                         material: materials.add(StandardMaterial::from(color)),
                         transform: Transform::from_xyz(0.0, 0.0, 0.0),
                         ..default()
                     },
                     // RaycastPickTarget::default(), // <- Needed for the raycast backend.
                     // PickableBundle::default() // <- This one too
    ));
    commands.spawn(( // vertex
                     PbrBundle {
                         mesh: meshes.add(Mesh::from(shape::Cube { size: 0.025 })),
                         material: materials.add(Color::rgb(1.0, 1.0, 1.0).into()),
                         transform: Transform::from_xyz(origin.to_xyz().x.into_f32(), origin.to_xyz().y.into_f32(), origin.to_xyz().z.into_f32()),
                         ..default()
                     },
                     // RaycastPickTarget::default(), // <- Needed for the raycast backend.
                     // PickableBundle::default() // <- This one too
    ));
}


/*


// Root node
        commands.spawn(NodeBundle {
            style: Style {
                flex_direction: FlexDirection::Column,
                justify_content: JustifyContent::Center,
                align_items: AlignItems::Center,
                ..Default::default()
            },
            ..Default::default()
        }).with_children(|parent| {
            // Text node
            parent.spawn(TextBundle {
                text: Text::from_section(
                    "This is some text",
                    TextStyle {
                        font: asset_server.load("fonts/FiraMono-Medium.ttf"),
                        font_size: 40.0,
                        color: Color::WHITE,
                    },
                ),
                ..Default::default()
            });

            // Checkbox (custom implementation)
            parent.spawn(NodeBundle {
                style: Style {
                    justify_content: JustifyContent::Center,
                    align_items: AlignItems::Center,
                    ..Default::default()
                },
                ..Default::default()
            }).with_children(|parent| {
                parent.spawn(TextBundle {
                    text: Text::from_section(
                        "X", // Simulating a checked state
                        TextStyle {
                            font: asset_server.load("fonts/FiraSans-Bold.ttf"),
                            font_size: 20.0,
                            color: Color::BLACK,
                        },
                    ),
                    ..Default::default()
                });
            });
        });
 */